#! /usr/bin/env python
#
# ======================================================================== 
#
# Python module for SBIG CCD camera - python test code
# Copyright (C) 2006 Laurent Le Guillou <llg@lpnhep.in2p3.fr>
# Mercator Telescope, C/A. de Abreu 70, 38700 Santa Cruz de La Palma, Spain
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301  USA
#
# ========================================================================

import os
import sys
import time
import sbig                   # this module
import pyfits                 # only for FITS output

# === Opening the driver

print "Open the driver..."

try:
    sbig.open_driver()
    devices = sbig.query_usb() # return list of USB SBIG cameras found
    # There is a bug with this last function
except sbig.SBIGDriverError, msg:
    print >>sys.stderr, msg
    sys.exit(2)

# === Opening the device

print "Open the device..."

try:
    sbig.open_device(sbig.DEV_USB)
except sbig.SBIGDriverError, msg:
    print >>sys.stderr, "Unable to open SBIG camera device"
    print >>sys.stderr, "Error: ", msg
    sys.exit(3)

# === Establish link with the device
#
# it is often needed to repeat establish_link until it succeed.
# (Ref: Jan Soldan)

print "Establish the link with the camera..."

tries=5
camera_type = sbig.NO_CAMERA
for i in xrange(tries):
    try:
        camera_type = sbig.establish_link()
        break
    except sbig.SBIGDriverError, msg:
        print >>sys.stderr, msg, "Trying again..."
        pass
if (camera_type == sbig.NO_CAMERA):
    print >>sys.stderr, "Cannot establish link with the camera."
    sys.exit(4)

# === Get infos about the CCD and the available readout modes 
#

print "Get CCD infos..."

ccdinfo = sbig.get_ccd_info()
print "-----------------------------------------------------"
print "Readout Modes: ", len(ccdinfo['modes'])
modes_id = ccdinfo['modes'].keys()
modes_id.sort()
for m in modes_id:
    print "--- mode #"+str(m)
    print "       width",        ccdinfo['modes'][m]['width'],        "pixel"
    print "       height",       ccdinfo['modes'][m]['height'],       "pixel"
    print "       pixel width",  ccdinfo['modes'][m]['pixel_width'],  "micron"
    print "       pixel height", ccdinfo['modes'][m]['pixel_height'], "micron"
    print "       gain",         ccdinfo['modes'][m]['gain'],         "e-/ADU"

# === Read all the control parameters (DCP_*)
# Just to check

print "Get control parameters..."

parameter_names = [n for n in dir(sbig) if n[0:3]=='DCP']
for p in parameter_names:
    print "    ", p, sbig.get_driver_control(eval('sbig.' + p))

# Test only (useless)
# sbig.set_driver_control(sbig.DCP_USB_FIFO_SIZE, 8192) # divide by 2
# sbig.set_driver_control(sbig.DCP_USB_FIFO_SIZE, 4096) # divide by 2

# sbig.set_driver_control(sbig.DCP_USB_FIFO_ENABLE, 1) 
# sbig.set_driver_control(sbig.DCP_USB_FIFO_ENABLE, 0) 

# Read all the control parameters (DCP_*)
# parameter_names = [n for n in dir(sbig) if n[0:3]=='DCP']
# for p in parameter_names:
#     print "    ", p, sbig.get_driver_control(eval('sbig.' + p))


# === Temperature regulation
#
print "Set temperature..."

# Enable temperature regulation at 15 C
setpoint = 15.0
sbig.set_temperature_regulation(sbig.REGULATION_ON, setpoint_celsius=setpoint)

print "Wait until the temperature has reached the setpoint..."

# Wait until the temperature reach the setpoint value
while True:
    temp_status = sbig.query_temperature_status()
    print "ccd_celsius = ", temp_status['ccd_celsius']
    print "setpoint = ", temp_status['setpoint_celsius']
    if abs(temp_status['ccd_celsius'] -
           temp_status['setpoint_celsius']) < 1.0:
        break
    time.sleep(1)

# Regulation will be "frozen" during readout to minimize readout noise.
# (Freeze will start at the first readout_line command and stop
# when the end_readout command will be sent.)
sbig.set_temperature_regulation(sbig.REGULATION_ENABLE_AUTOFREEZE)


# === Exposure
#

exposure = 10.0

print "Starting a %2.1fs exposure..." % exposure

# Start exposure (10.0s)
sbig.start_exposure(ccd=sbig.CCD_IMAGING, exposure=exposure)

# Waiting until the end of the exposure (by checking status)
# sbig.query_command_status() gives status of the last command
done = 0.0             # just for display
wait_increment = 1.0   # just for display
while (sbig.query_command_status() == sbig.STATUS_INTEGRATING):
    time.sleep(wait_increment)
    done += wait_increment
    print "%2.1f done..." % done

print "Ending exposure..."
# ending exposure (can also be used to end it prematurely)
sbig.end_exposure(ccd=sbig.CCD_IMAGING)

# === Readout


mode=0  # readout mode (binning 1x1 i.e. not binned)
fullheight = ccdinfo['modes'][mode]['height']
fullwidth = ccdinfo['modes'][mode]['width']
height=fullheight
width=fullwidth
top=0
left=0
# height=400
# width=500
# top=100
# left=150

print "Starting readout in mode", mode, "..."

time_start = time.time()

sbig.start_readout(ccd=sbig.CCD_IMAGING, mode=mode,
                   top=top, left=left,
                   height=height, width=width)

# You may read line by line with readout_line(...),
# or a block with readout_lines(...)
# image is a numarray 2D-array.

image = sbig.readout_lines(ccd=sbig.CCD_IMAGING, mode=mode,
                           start=left, length=width, lines=height)

# Alternative: line by line
# be careful, if an exception occured, you can be sure
# that the resulting image is corrupted
# image = numarray.zeros(shape=(height, width))
# for line in xrange(height):
#     # you should repeat the next line to read every row
#     tries = 5
#     while (tries > 0):
#         try:
#             l = sbig.readout_line(ccd=sbig.CCD_IMAGING, mode=0,
#                                   start=0, length=width)
#             break
#         except sbig.SBIGDriverError, msg:
#             print >>sys.stderr, ("line [%d]" % line), msg
#             tries -= 1
#             pass

#     if (tries == 0):
#         print >>sys.stderr, "error: timeout (5 times)"
#         sys.exit(2)

#     image[line] = l

print "Readout complete."
print "Image stats:"
print "    min = ", image.min(), "  max = ", image.max()
print "    mean = ", image.mean(), "  rms = ", image.stddev()

sbig.end_readout(ccd=sbig.CCD_IMAGING)

time_end = time.time()

print "Readout in %f s" % (time_end - time_start)

# === Saving the image as a FITS image

print "Writing image as 'out.fits'..."

filename = "out.fits"
fitsobj = pyfits.HDUList()
hdu = pyfits.PrimaryHDU()
# scale the data to Int16 with user specified bscale/bzero
# no other operation should manipulate the data after this
# hdu.data = image
hdu.data = image*1.0  # seems needed for the next operation
hdu.scale('Int16', '', bzero=32768)
fitsobj.append(hdu)
try:
    os.unlink(filename)
except OSError:
    pass
fitsobj.writeto(filename)

print "Writing image done."

# === Closing the device and the driver when the program ends

print "Closing device and driver..."

sbig.close_device()
sbig.close_driver()

print "Bye."

