#! /usr/bin/env python
#
# ======================================================================== 
#
# Python module for SBIG CCD camera - python test code
# 27/08/2007 Bart Vandenbussche 
#
# Example on using the miscellaneous_control method to keep the shutter open
# between exposures
# ========================================================================

import sys
import os
import time
import sbig
import numarray
import pyfits

sbig.open_driver()
sbig.open_device(sbig.DEV_USB)      

tries=5
camera_type = sbig.NO_CAMERA
for i in xrange(tries):
    try:
        camera_type = sbig.establish_link()
        break
    except sbig.SBIGDriverError:
        pass

if (camera_type == sbig.NO_CAMERA):
    print >>sys.stderr, "Cannot establish link with the camera."
    sys.exit(2)


ccdinfo = sbig.get_ccd_info()


# ================ First exposure: using shutter ==============

exposure = 5; # 5 seconds

sbig.start_exposure(ccd=sbig.CCD_IMAGING, exposure=exposure, shutter=sbig.SC_OPEN_SHUTTER)
while (sbig.query_command_status() == sbig.STATUS_INTEGRATING):
    time.sleep(1)
    sbig.end_exposure(ccd=sbig.CCD_IMAGING)
        
mode = 0 # no binning
fullheight = ccdinfo['modes'][mode]['height']
fullwidth = ccdinfo['modes'][mode]['width']
height = fullheight
width = fullwidth
top = 0
left = 0
    
sbig.start_readout(ccd=sbig.CCD_IMAGING, mode=mode, top=top, left=left,
                       height=height, width=width)
im_shut = sbig.readout_lines(ccd=sbig.CCD_IMAGING, mode=mode,
                              start=left, length=width, lines=height);
sbig.end_readout(ccd=sbig.CCD_IMAGING)
    
print im_shut.shape
print "Shutter open/close"
print "  min = ", im_shut.min(), "  max = ", im_shut.max()
print "  mean = ", im_shut.mean(), "  rms = ", im_shut.stddev()
    
filename = "with_shutter.fits" 
fitsobj = pyfits.HDUList()
hdu = pyfits.PrimaryHDU()
# scale the data to Int16 with user specified bscale/bzero
# no other operation should manipulate the data after this
# hdu.data = dark
hdu.data = im_shut*1.0  # seems needed for the next operation
hdu.scale('Int16', '', bzero=32768)
fitsobj.append(hdu)
fitsobj.writeto(filename)



# ================ Second exposure: keeping the shutter open  ==============

# miscellaneous_control parameters:
# fan_enable:
#    False : disable fan
#    True  : enable fan
# shutter_command: 
#    sbig.SC_LEAVE_SHUTTER      : leave shutter alone
#    sbig.SC_OPEN_SHUTTER       : open shutter
#    sbig.SC_CLOSE_SHUTTER      : close shutter
#    sbig.SC_INITIALIZE_SHUTTER : reinitialize shutter, 
#    sbig.SC_OPEN_EXT_SHUTTER   : open ST-L external shutter
#    sbig.SC_CLOSE_EXT_SHUTTER  : close ST-L external shutter
# led_state 
#    sbig.LED_OFF               : LED off
#    sbig.LED_ON                : LED on
#    sbig.LED_BLINK_LOW         : LED blink at low rate
#    sbig.LED_BLINK_HIGH        : LED blink at high rate
# 

# Example: Open the shutter, switch off LED, leave fan on:

sbig.miscellaneous_control(fan_enable=1, shutter_command=sbig.SC_OPEN_SHUTTER, led_state=sbig.LED_OFF)


exposure = 5; # 5 seconds

sbig.start_exposure(ccd=sbig.CCD_IMAGING, exposure=exposure, shutter=sbig.SC_LEAVE_SHUTTER)
while (sbig.query_command_status() == sbig.STATUS_INTEGRATING):
    time.sleep(1)
    sbig.end_exposure(ccd=sbig.CCD_IMAGING)
        
mode = 0 # no binning
fullheight = ccdinfo['modes'][mode]['height']
fullwidth = ccdinfo['modes'][mode]['width']
height = fullheight
width = fullwidth
top = 0
left = 0
    
sbig.start_readout(ccd=sbig.CCD_IMAGING, mode=mode, top=top, left=left,
                       height=height, width=width)
im_noshut = sbig.readout_lines(ccd=sbig.CCD_IMAGING, mode=mode,
                              start=left, length=width, lines=height);
sbig.end_readout(ccd=sbig.CCD_IMAGING)
    
print im_shut.shape
print "Shutter open/close"
print "  min = ", im_noshut.min(), "  max = ", im_noshut.max()
print "  mean = ", im_noshut.mean(), "  rms = ", im_noshut.stddev()
    
filename = "without_shutter.fits" 
fitsobj = pyfits.HDUList()
hdu = pyfits.PrimaryHDU()
# scale the data to Int16 with user specified bscale/bzero
# no other operation should manipulate the data after this
# hdu.data = dark
hdu.data = im_noshut*1.0  # seems needed for the next operation
hdu.scale('Int16', '', bzero=32768)
fitsobj.append(hdu)
fitsobj.writeto(filename)

sbig.miscellaneous_control(fan_enable=1, shutter_command=sbig.SC_OPEN_SHUTTER, led_state=sbig.LED_OFF)

# ==================================================================


sbig.close_device()
sbig.close_driver()


