/* ======================================================================== *
 * 
 * HERMES Project
 * SBIG Python low-level module
 *
 * Author: Laurent Le Guillou <laurentl@ster.kuleuven.be>
 * Updates by Bart Vandenbussche <bart@ster.kuleuven.be>
 * ======================================================================== *
 *
 * Python module for SBIG CCD camera - version 0.8
 * Copyright (C) 2006 Laurent Le Guillou <llg@lpnhep.in2p3.fr>
 * Mercator Telescope, C/A. de Abreu 70, 38700 Santa Cruz de La Palma, Spain
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301  USA
 *
 * ======================================================================== *
 *
 * $Id: sbig.c,v 1.8 2007/08/27 14:50:02 bart Exp $
 *
 * ======================================================================== *
 *
 * $Log: sbig.c,v $
 * Revision 1.8  2007/08/27 14:50:02  bart
 * Added miscellaneous_control method to manipulate the shutter, led and fan
 *
 * Revision 1.7  2006/10/18 09:43:03  laurentl
 * *** empty log message ***
 *
 * Revision 1.6  2006/07/26 11:23:44  laurentl
 * *** empty log message ***
 *
 * Revision 1.5  2006/07/26 10:53:19  laurentl
 * *** empty log message ***
 *
 * Revision 1.4  2006/07/18 13:53:37  laurentl
 * *** empty log message ***
 *
 * Revision 1.3  2006/07/13 17:59:45  laurentl
 * Added shutter constants and udev rules
 *
 * Revision 1.2  2006/07/11 01:55:27  laurentl
 * Added version __version__
 *
 * Revision 1.1  2006/07/11 00:11:18  laurentl
 * Imported sbig python wrapper
 *
 *
 * ======================================================================== */

#define VERSION "0.9"

/* ======================================================================== *
 * Sample python code using this module (very preliminary code)             *
 *

 import sys
 import sbig
 import pyfits                 # only for FITS output

 sbig.open_driver()
 sbig.query_usb()              # return list of USB cameras
                               # [TODO: BUG: works only once]

 sbig.open_device(sbig.DEV_USB)      

 # it is often needed to repeat establish_link until it succeed.
 # (Ref: Jan Soldan)
 tries=5
 camera_type = sbig.NO_CAMERA
 for i in xrange(tries): 
     try:
         camera_type = sbig.establish_link()
         break
     except sbig.SBIGDriverError, msg:
         print >>sys.stderr, msg, "Trying again..."
         pass

 if (camera_type == sbig.NO_CAMERA):
     print >>sys.stderr, "Cannot establish link with the camera."
     sys.exit(2)

 ccdinfo = sbig.get_ccd_info()
 print ccdinfo

 # Read all the control parameters (DCP_*)
 parameter_names = [n for n in dir(sbig) if n[0:3]=='DCP']
 for p in parameter_names:
     print p, sbig.get_driver_control(eval('sbig.' + p))

 # Enable temperature regulation at 15C
 sbig.set_temperature_regulation(sbig.REGULATION_ON, setpoint_celsius=15.0)

 # Regulation will be "frozen" during readout to minimize readout noise.
 # (Freeze will start at the first readout_line command and stop
 # when the end_readout command will be sent.)
 sbig.set_temperature_regulation(sbig.REGULATION_ENABLE_AUTOFREEZE)
 
 # Start exposure (10.0s)
 sbig.start_exposure(ccd=sbig.CCD_IMAGING, exposure=10.0)

 # Waiting until the end of the exposure (by checking status)
 # sbig.query_command_status() gives status of the last command
 while (sbig.query_command_status() == sbig.STATUS_INTEGRATING):
     time.sleep(.1)

 # ending exposure (can also be used to end it prematurely)
 sbig.end_exposure(ccd=sbig.CCD_IMAGING)

 mode=0  # readout mode (binning 1x1 i.e. not binned)
 fullheight = ccdinfo['modes'][mode]['height']
 fullwidth = ccdinfo['modes'][mode]['width']
 height=fullheight
 width=fullwidth
 top=0
 left=0

 sbig.start_readout(ccd=sbig.CCD_IMAGING, mode=mode, 
                    top=top, left=left,
                    height=height, width=width)

 # You may read line by line with readout_line(...),
 # or a block with readout_lines(...)
 # image is a numarray 2D-array.

 image = sbig.readout_lines(ccd=sbig.CCD_IMAGING, mode=mode, 
                            start=left, length=width, lines=height)

 sbig.end_readout(ccd=sbig.CCD_IMAGING)

 sbig.close_device()
 sbig.close_driver()

 filename = "out.fits"
 fitsobj = pyfits.HDUList()
 hdu = pyfits.PrimaryHDU()
 hdu.data = image
 fitsobj.append(hdu)
 try:
     os.unlink(filename)
 except OSError:
     pass
 fitsobj.writeto(filename)

 * ======================================================================== */

#include <Python.h>
#include <numarray/libnumarray.h>

#include <math.h>

#include <sbigudrv.h>


/* ======================================================================== *
 * Conversion BCD <-> decimal
 *
 */

long bcd(long dec) 
{
  int i = 0;
  long result = 0;
  while (dec) {
    result += (dec % 10) << (4 * i);
    dec /= 10;
    i++;
  }
  return(result);
}

long dec(long bcd)
{
  int i = 1;
  long result = 0;
  while (bcd) {
    result += bcd % 16 * i;
    i *= 10;
    bcd /= 16;
  }
  return(result);
}

/* ======================================================================== *
 * 
 * Exception definitions and Error messages
 *
 */

static PyObject *sbig_error;

#define INTERNAL_BUFFER_LENGTH 64
char internal_buffer[INTERNAL_BUFFER_LENGTH];

char * sbig_error_string(PAR_ERROR err)
{
  GetErrorStringParams gsp;
  GetErrorStringResults gsr;

  gsp.errorNo = err;
  SBIGUnivDrvCommand(CC_GET_ERROR_STRING, &gsp, &gsr);
  strncpy(internal_buffer, gsr.errorString, INTERNAL_BUFFER_LENGTH-1);
  internal_buffer[INTERNAL_BUFFER_LENGTH-1]='\0';
  return internal_buffer;
}

/* last command id (for query_command_status) */
short sbig_last_command;

/* ======================================================================== *
 * 
 * (3.1) Driver related commands [Driver documentation, section 3.1]
 *
 * open_driver()
 * close_driver()
 * open_device(device, lpt_base_address=None, ip_address=None)
 * close_device()
 * get_driver_info(request=0)
 * set_driver_handle(handle)
 * get_driver_handle()
 * 
 *
 */

/* ----- open_driver ------------------------------------------------------ */

static PyObject *
sbig_open_driver(PyObject *self, PyObject *args)
{
  unsigned short sts;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }
  
  sbig_last_command = CC_OPEN_DRIVER;
  sts = SBIGUnivDrvCommand(CC_OPEN_DRIVER, NULL, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- close_driver ----------------------------------------------------- */

static PyObject *
sbig_close_driver(PyObject *self, PyObject *args)
{
  short sts;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }

  sbig_last_command = CC_CLOSE_DRIVER;
  sts = SBIGUnivDrvCommand(CC_CLOSE_DRIVER, NULL, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- open_device ------------------------------------------------------ */

static PyObject *
sbig_open_device(PyObject *self, PyObject *args, PyObject *keywds)
{
  short sts;
  OpenDeviceParams gsp;
  gsp.deviceType=DEV_USB;   /* USB device by default */
  gsp.lptBaseAddress=0x378; /* LPT1 */
  gsp.ipAddress=0;

  static char *kwlist[] = {"device", 
			   "lpt_base_address", 
			   "ip_address", 
			   NULL};

  if (!PyArg_ParseTupleAndKeywords(args, keywds, "|iii", kwlist, 
				   &gsp.deviceType, 
				   &gsp.lptBaseAddress,
				   &gsp.ipAddress)) {
    return NULL; 
  }

  sbig_last_command = CC_OPEN_DEVICE;
  sts = SBIGUnivDrvCommand(CC_OPEN_DEVICE, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- close_device ----------------------------------------------------- */

static PyObject *
sbig_close_device(PyObject *self, PyObject *args)
{
  short sts;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }

  sbig_last_command = CC_CLOSE_DEVICE;
  sts = SBIGUnivDrvCommand(CC_CLOSE_DEVICE, NULL, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- get_driver_info -------------------------------------------------- */

static PyObject *
sbig_get_driver_info(PyObject *self, PyObject *args, PyObject *keywds)
{
  short sts;
  short request=0;
  char version[16];
  GetDriverInfoParams gsp;
  GetDriverInfoResults0 gsr;

  static char *kwlist[] = {"request", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, keywds, "|i", kwlist, &request)) {
    return NULL; 
  }

  gsp.request = request;
  sbig_last_command = CC_GET_DRIVER_INFO;
  sts = SBIGUnivDrvCommand(CC_GET_DRIVER_INFO, &gsp, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  /* Version is a BCD representation XX.XX (cf. documentation) */
  sprintf(version, "%02x.%02x", 
	  ((gsr.version & 0xFF00) >> 8),
	  (gsr.version & 0x00FF));

  return Py_BuildValue("{s:s,s:s,s:i}", 
		       "version",                version, 
		       "name",                   gsr.name, 
		       "max_request",            gsr.maxRequest);
}

/* ----- get_driver_handle ------------------------------------------------ */

static PyObject *
sbig_get_driver_handle(PyObject *self, PyObject *args)
{
  short sts;
  GetDriverHandleResults gsr;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }

  sbig_last_command = CC_GET_DRIVER_HANDLE;
  sts = SBIGUnivDrvCommand(CC_GET_DRIVER_HANDLE, NULL, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  return Py_BuildValue("i", gsr.handle);
}

/* ----- set_driver_handle ------------------------------------------------ */

/* MISSING: TODO:
 *    Manage INVALID_HANDLE_VALUE
 */

static PyObject *
sbig_set_driver_handle(PyObject *self, PyObject *args, PyObject *keywds)
{
  short sts;
  SetDriverHandleParams gsp;

  static char *kwlist[] = {"handle", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &gsp.handle)) {
    return NULL; 
  }

  sbig_last_command = CC_SET_DRIVER_HANDLE;
  sts = SBIGUnivDrvCommand(CC_SET_DRIVER_HANDLE, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

/* ======================================================================== *
 * 
 * (3.2) Exposure related commands [Driver documentation, section 3.2]
 *
 * start_exposure(ccd, exposure, antiblooming, shutter)
 * end_exposure(ccd)
 * start_readout(ccd, mode, top, left, height, width)
 * readout_line(ccd, mode, start, length)
 * dump_lines(ccd, mode, length)
 * end_readout(ccd)
 * 
 *
 */

/* ----- start_exposure --------------------------------------------------- */

static PyObject *
sbig_start_exposure(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  StartExposureParams gsp;
  float exposure_seconds;

  gsp.abgState = ABG_LOW7;
  gsp.openShutter = SC_OPEN_SHUTTER;
  
  static char *kwlist[] = {"ccd", "exposure", "antiblooming", "shutter",  
			   NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "if|ii", kwlist, 
				   &gsp.ccd, &exposure_seconds, 
				   &gsp.abgState, &gsp.openShutter)) {
    return NULL; 
  }
  
  /* fprintf(stderr, "[%f][%ld]\n", exposure_seconds, gsp.exposureTime); */
  if (((gsp.ccd != CCD_IMAGING) && (exposure_seconds > 655.35)) ||
      (exposure_seconds > 167777.16)) {
    /* Exposure too long, will be converted in millisecs... Bad */
    PyErr_SetString(PyExc_ValueError, "exposure time too long");
    return NULL;
  }
  
  gsp.exposureTime = (long)floor(exposure_seconds * 100);

  sbig_last_command = CC_START_EXPOSURE;
  sts = SBIGUnivDrvCommand(CC_START_EXPOSURE, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- end_exposure ----------------------------------------------------- */

static PyObject *
sbig_end_exposure(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  EndExposureParams gsp;
  
  static char *kwlist[] = {"ccd", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &gsp.ccd)) {
    return NULL; 
  }
  
  sbig_last_command = CC_END_EXPOSURE;
  sts = SBIGUnivDrvCommand(CC_END_EXPOSURE, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- start_readout ---------------------------------------------------- */

static PyObject *
sbig_start_readout(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  StartReadoutParams gsp;
  gsp.readoutMode = 0;

  static char *kwlist[] = {"ccd", "mode", 
			   "top", "left", "height", "width", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iiiiii", kwlist, 
				   &gsp.ccd, &gsp.readoutMode, 
				   &gsp.top, &gsp.left, 
				   &gsp.height, &gsp.width)) {
    return NULL; 
  }
  
  sbig_last_command = CC_START_READOUT;
  sts = SBIGUnivDrvCommand(CC_START_READOUT, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}


/* ----- readout_line ----------------------------------------------------- */

static PyObject *
sbig_readout_line(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  ReadoutLineParams gsp;
  gsp.readoutMode = 0;
  unsigned short * read_array;

  static char *kwlist[] = {"ccd", "mode", 
			   "start", "length", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iiii", kwlist, 
				   &gsp.ccd, &gsp.readoutMode, 
				   &gsp.pixelStart, &gsp.pixelLength)) {
    return NULL; 
  }
  
  if (gsp.pixelLength <= 0) {
    PyErr_SetString(PyExc_ValueError, "invalid length value");
    return NULL;
  }

  read_array = (unsigned short *)malloc(gsp.pixelLength * sizeof(short));
  if (read_array == NULL) {
    PyErr_NoMemory();
      /* PyErr_SetString((PyExc_MemoryError, "out of memory"); */
    return NULL;
  }

  sbig_last_command = CC_READOUT_LINE;
  sts = SBIGUnivDrvCommand(CC_READOUT_LINE, &gsp, read_array);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    free(read_array);
    return NULL;
  }

  return Py_BuildValue("O", NA_NewArray((void *)read_array, tUInt16, 
					1, gsp.pixelLength));

}

/* ----- readout_lines ---------------------------------------------------- */
/*
 * Read several lines at once (faster than repeating the previous 
 * function, due to numarray memory copies).
 * [Does not seems to give a great gain in readout time...]
 *
 */

static PyObject *
sbig_readout_lines(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  ReadoutLineParams gsp;
  gsp.readoutMode = 0;
  unsigned short * read_array;
  int lines = 0;
  int line;

  static char *kwlist[] = {"ccd", "mode", 
			   "start", "length", "lines", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iiiii", kwlist, 
				   &gsp.ccd, &gsp.readoutMode, 
				   &gsp.pixelStart, &gsp.pixelLength,
				   &lines)) {
    return NULL; 
  }
  
  if (gsp.pixelLength <= 0) {
    PyErr_SetString(PyExc_ValueError, "invalid length value");
    return NULL;
  }
  if (lines <= 0) {
    PyErr_SetString(PyExc_ValueError, "invalid lines value");
    return NULL;
  }

  read_array = (unsigned short *)malloc(gsp.pixelLength * lines * 
					sizeof(short));
  if (read_array == NULL) {
    PyErr_NoMemory();
    return NULL;
  }

  for(line=0; line<lines; line++) {
    sbig_last_command = CC_READOUT_LINE;
    sts = SBIGUnivDrvCommand(CC_READOUT_LINE, &gsp, 
			     (read_array + line * gsp.pixelLength));
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      free(read_array);
      return NULL;
    }
  }

  return Py_BuildValue("O", NA_NewArray((void *)read_array, tUInt16, 
					2, lines, gsp.pixelLength));
  
}



/* ----- dump_lines ------------------------------------------------------- */

static PyObject *
sbig_dump_lines(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  DumpLinesParams gsp;
  
  static char *kwlist[] = {"ccd", "mode", "lines", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iii", kwlist, 
				   &gsp.ccd, &gsp.readoutMode, 
				   &gsp.lineLength)) {
    return NULL; 
  }
  
  sbig_last_command = CC_DUMP_LINES;
  sts = SBIGUnivDrvCommand(CC_DUMP_LINES, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- end_readout ------------------------------------------------------ */

static PyObject *
sbig_end_readout(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  EndReadoutParams gsp;

  static char *kwlist[] = {"ccd", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &gsp.ccd)) {
    return NULL; 
  }
  
  sbig_last_command = CC_END_READOUT;
  sts = SBIGUnivDrvCommand(CC_END_READOUT, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

/* ======================================================================== *
 * 
 * (3.3) Temperature related commands [Driver documentation, section 3.3]
 *
 * set_temperature_regulation(regulation, ...)
 * query_temperature_status()
 * 
 *
 */

/* ----- temperature units conversions ------------------------------------ *
 *
 * Temperature A/D <-> Celsius conversions 
 *
 */

#define  T0                 25.0
#define  R0                  3.0

#define MAX_AD            4096.0

#define R_RATIO_CCD          2.57
#define R_BRIDGE_CCD        10.0
#define DT_CCD              25.0

#define R_RATIO_AMBIENT      7.791
#define R_BRIDGE_AMBIENT     3.0
#define DT_AMBIENT          45.0

float AD_to_Celsius(int AD, float r_ratio, float r_bridge, float dt) {
  float r;
  r = r_bridge / (MAX_AD/AD - 1.0);
  return (T0 - dt * (logf(r/R0) / logf(r_ratio)));
}

int Celsius_to_AD(float Celsius, float r_ratio, float r_bridge, float dt) {
  float r;
  r = R0 * expf(logf(r_ratio) * (T0 - Celsius) / dt);
  return (int)floorf((MAX_AD / (r_bridge/r + 1.0)) + .5);  /* round equiv. */
}


/* ----- set_temperature_regulation --------------------------------------- */


static PyObject *
sbig_set_temperature_regulation(PyObject *self, 
				PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  SetTemperatureRegulationParams gsp;
  QueryTemperatureStatusResults gsr;

  int int_unset = -1;
  float float_unset = -1.0e9;

  float setpoint_celsius;
  int setpoint_ad;
  int power;

  setpoint_celsius = float_unset;
  setpoint_ad = int_unset;
  power = int_unset;

  gsp.regulation = 0;

  static char *kwlist[] = {"regulation", 
			   "setpoint_celsius", 
			   "setpoint_ad", 
			   "power",
			   NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "i|fii", kwlist, 
				   &gsp.regulation, 
				   &setpoint_celsius, &setpoint_ad,
				   &power)) {
    return NULL; 
  }

  switch(gsp.regulation) {
  case REGULATION_ON:
    /* a valid setpoint (either ad or celsius) is needed */
    if (setpoint_celsius == float_unset) {
      if (setpoint_ad == int_unset) {
	sbig_last_command = CC_QUERY_TEMPERATURE_STATUS;
	sts = SBIGUnivDrvCommand(CC_QUERY_TEMPERATURE_STATUS, NULL, &gsr);
	if (sts != 0) {
	  PyErr_SetString(sbig_error, sbig_error_string(sts));
	  return NULL;
	}
	setpoint_ad = gsr.ccdSetpoint;
      }
    } else {
      if (setpoint_ad != int_unset) {  /* both are set */
	PyErr_SetString(PyExc_ValueError, "You should provide only one of the (setpoint_celsius, setpoint_ad) parameters, not both.");
	return NULL;
      } else {
	setpoint_ad = Celsius_to_AD(setpoint_celsius,
				    R_RATIO_CCD, R_BRIDGE_CCD, DT_CCD);
      }
    } 

    gsp.ccdSetpoint = setpoint_ad;
    sbig_last_command = CC_SET_TEMPERATURE_REGULATION;
    sts = SBIGUnivDrvCommand(CC_SET_TEMPERATURE_REGULATION, &gsp, NULL);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    break;

  case REGULATION_OFF:
  case REGULATION_FREEZE:
  case REGULATION_UNFREEZE:
  case REGULATION_ENABLE_AUTOFREEZE:
  case REGULATION_DISABLE_AUTOFREEZE:
    /* for those cases, not other parameter is needed */
    sbig_last_command = CC_QUERY_TEMPERATURE_STATUS;
    sts = SBIGUnivDrvCommand(CC_QUERY_TEMPERATURE_STATUS, NULL, &gsr);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    setpoint_ad = gsr.ccdSetpoint;

    gsp.ccdSetpoint = setpoint_ad;
    sbig_last_command = CC_SET_TEMPERATURE_REGULATION;
    sts = SBIGUnivDrvCommand(CC_SET_TEMPERATURE_REGULATION, &gsp, NULL);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    break;
  
  case REGULATION_OVERRIDE:
    /* a valid TE power (0-255) is needed */
    if (power == int_unset) {
      PyErr_SetString(PyExc_ValueError, "You should provide a valid TE power value (0-255) parameter.");
      return NULL;
    }
    gsp.ccdSetpoint = power;
    sbig_last_command = CC_SET_TEMPERATURE_REGULATION;
    sts = SBIGUnivDrvCommand(CC_SET_TEMPERATURE_REGULATION, &gsp, NULL);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    break;
    
  default:
    PyErr_SetString(PyExc_ValueError, "invalid regulation value");
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}


/* ----- query_temperature_status ----------------------------------------- */

static PyObject *
sbig_query_temperature_status(PyObject *self, PyObject *args)
{
  unsigned short sts;
  QueryTemperatureStatusResults gsr;
  float setpoint_celsius, ccd_celsius, ambient_celsius;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }
  
  sbig_last_command = CC_QUERY_TEMPERATURE_STATUS;
  sts = SBIGUnivDrvCommand(CC_QUERY_TEMPERATURE_STATUS, NULL, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  ccd_celsius = AD_to_Celsius(gsr.ccdThermistor, 
			      R_RATIO_CCD, R_BRIDGE_CCD, DT_CCD);
  setpoint_celsius = AD_to_Celsius(gsr.ccdSetpoint, 
				   R_RATIO_CCD, R_BRIDGE_CCD, DT_CCD);
  ambient_celsius = AD_to_Celsius(gsr.ambientThermistor, 
				  R_RATIO_AMBIENT, 
				  R_BRIDGE_AMBIENT, DT_AMBIENT);

  return Py_BuildValue("{s:i,s:i,s:i,s:f,s:i,s:f,s:i,s:f,s:i}", 
		       "enabled",          (gsr.enabled ? 1 : 0), 
		       "frozen",           ((gsr.enabled & 
					    REGULATION_FROZEN_MASK) ? 1 : 0),
		       "setpoint_ad",      gsr.ccdSetpoint,
		       "setpoint_celsius", setpoint_celsius,
		       "ccd_ad",           gsr.ccdThermistor,
		       "ccd_celsius",      ccd_celsius,
		       "ambient_ad",       gsr.ambientThermistor,
		       "ambient_celsius",  ambient_celsius,
		       "power",            gsr.power
		       );

}

/* ======================================================================== *
 * 
 * (3.5) General purpose commands [Driver documentation, section 3.5]
 *
 * establish_link()
 * get_ccd_info()
 * query_command_status(command)
 * miscellaneous_control(fan_enable=TRUE|FALSE, shutter_command=0..5, led_state=0..3)
 * get_link_status()
 * set_driver_control(control_parameter, control_value)
 * get_driver_control(control_parameter)
 * query_usb()
 * [...]
 *
 */

/* ----- establish_link --------------------------------------------------- */

static PyObject *
sbig_establish_link(PyObject *self, PyObject *args)
{
  unsigned short sts;
  EstablishLinkParams gsp;
  EstablishLinkResults gsr;
  gsp.sbigUseOnly = 0;
  
  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }
  
  sbig_last_command = CC_ESTABLISH_LINK;
  sts = SBIGUnivDrvCommand(CC_ESTABLISH_LINK, &gsp, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  return Py_BuildValue("i", gsr.cameraType);
}

/* ----- get_ccd_info ----------------------------------------------------- */

static PyObject *
sbig_get_ccd_info(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  int mode, i;
  GetCCDInfoParams gsp;
  GetCCDInfoResults0 gsr0;
  GetCCDInfoResults2 gsr2;
  GetCCDInfoResults3 gsr3;
  GetCCDInfoResults4 gsr4;
  char firmware[16];

  PyObject * item;
  PyObject * modes;
  PyObject * badcolumns;
  PyObject * result;

  gsp.request = 0;

  static char *kwlist[] = {"request", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "|i", kwlist, &gsp.request)) {
    return NULL; 
  }

  switch(gsp.request) {
  case CCD_INFO_IMAGING:
  case CCD_INFO_TRACKING:
    sbig_last_command = CC_GET_CCD_INFO;
    sts = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &gsp, &gsr0);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    /* Firmware is coded as a BCD number XX.XX */
    sprintf(firmware, "%02x.%02x", 
	    ((gsr0.firmwareVersion & 0xFF00) >> 8),
	    (gsr0.firmwareVersion & 0x00FF));

    /* Now build the resulting structure as a Python dictionary */

    result = Py_BuildValue("{s:s,s:i,s:s}",
			   "firmware",   firmware,
			   "type",       gsr0.cameraType,
			   "name",       gsr0.name);

    /* 
       Build the mode list 'modes':
       'modes': { mode_id: {...}, mode_id: {...} }
       each mode is a dictionary {'width': ..., ...}
       to be able to do:
       height = result['modes'][mode_id]['height']
    */

    modes = Py_BuildValue("{}");

    for (mode=0 ; mode<gsr0.readoutModes ; mode++) {
      item = Py_BuildValue("{s:i,s:i,s:f,s:f,s:f}",
			   "width",        gsr0.readoutInfo[mode].width,
			   "height",       gsr0.readoutInfo[mode].height,
			   "gain",         
			   dec(gsr0.readoutInfo[mode].gain)/100.0,
			   "pixel_width",  
			   dec(gsr0.readoutInfo[mode].pixelWidth)/100.0,
			   "pixel_height", 
			   dec(gsr0.readoutInfo[mode].pixelHeight)/100.0);

      PyDict_SetItem(modes,
		     Py_BuildValue("i", gsr0.readoutInfo[mode].mode),
		     item);

    }

    /* Add the mode dictionary into the main result dictionary */

    PyDict_SetItem(result, Py_BuildValue("s", "modes"), modes);

    return result;

    break;

  case CCD_INFO_EXTENDED:
    sbig_last_command = CC_GET_CCD_INFO;
    sts = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &gsp, &gsr2);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    result = Py_BuildValue("{s:i,s:s}", 
			   "ABG",    gsr2.imagingABG,
			   "serial", gsr2.serialNumber);

    badcolumns = Py_BuildValue("()");
    for(i=0 ; i<gsr2.badColumns ; i++) {
      item = Py_BuildValue("i", gsr2.columns[i]);
      badcolumns = PySequence_InPlaceConcat(badcolumns, item);
    }

    PyDict_SetItem(result, Py_BuildValue("s", "bad_columns"), badcolumns);
    
    return result;

    break;

  case CCD_INFO_EXTENDED_5C:
    sbig_last_command = CC_GET_CCD_INFO;
    sts = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &gsp, &gsr3);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    result = Py_BuildValue("{s:i,s:i}",
			   "ad_size",     gsr3.adSize,
			   "filter_type", gsr3.filterType);
    return result;
    break;

  case CCD_INFO_EXTENDED2_IMAGING:
  case CCD_INFO_EXTENDED2_TRACKING:
    sbig_last_command = CC_GET_CCD_INFO;
    sts = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &gsp, &gsr4);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }
    result = Py_BuildValue("{s:i,s:i}",
			   "capabilities",  gsr4.capabilitiesBits,
			   "dump_extra",    gsr4.dumpExtra);
    return result;
    break;
  default:
    PyErr_SetString(PyExc_ValueError, "invalid request number");
    return NULL;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- query_command_status --------------------------------------------- *
 *
 * Give the status of the last driver command (using sbig_last_command)
 * It is possible to provide the command id (See CC_* constants in
 * the sbigudrv.h header). 
 *
 */

static PyObject *
sbig_query_command_status(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  QueryCommandStatusParams gsp;
  QueryCommandStatusResults gsr;

  gsp.command = sbig_last_command;

  static char *kwlist[] = {"command", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, keywds, "|i", kwlist, 
				   &gsp.command)) {
    return NULL; 
  }

  sts = SBIGUnivDrvCommand(CC_QUERY_COMMAND_STATUS, &gsp, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  return Py_BuildValue("i", gsr.status);
}

/* ----- miscellaneous control -------------------------------------------- */
static PyObject *
sbig_miscellaneous_control(PyObject *self,
		           PyObject *args, PyObject *keywds)
{
    unsigned short sts; 
    MiscellaneousControlParams gsp;

    int fan_enable;
    int shutter_command;
    int led_state;

    static char *kwlist[] = {"fan_enable",
                             "shutter_command",
                             "led_state", 
			     NULL}; 

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "iii", kwlist,
        &fan_enable, &shutter_command, &led_state)) {
      return NULL;
    }

    gsp.fanEnable = fan_enable;
    gsp.shutterCommand = shutter_command;
    gsp.ledState = led_state;

    sbig_last_command = CC_MISCELLANEOUS_CONTROL;

    sts = SBIGUnivDrvCommand(CC_MISCELLANEOUS_CONTROL, &gsp, NULL);
    if (sts != 0) {
      PyErr_SetString(sbig_error, sbig_error_string(sts));
      return NULL;
    }


    Py_INCREF(Py_None);
    return Py_None;
}


/* ----- read_offset ------------------------------------------------------ *
 *
 * Returns the CCD offset.
 *
 */

static PyObject *
sbig_read_offset(PyObject *self, PyObject *args, PyObject *keywds)
{
  unsigned short sts;
  ReadOffsetParams gsp;
  ReadOffsetResults gsr;

  static char *kwlist[] = {"ccd", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "i", kwlist, &gsp.ccd)) {
    return NULL; 
  }

  sbig_last_command = CC_READ_OFFSET;
  sts = SBIGUnivDrvCommand(CC_READ_OFFSET, &gsp, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  return Py_BuildValue("i", gsr.offset);

}

/* ----- get_link_status -------------------------------------------------- *
 *
 * Returns the status of the link established with the camera.
 *
 */

static PyObject *
sbig_get_link_status(PyObject *self, PyObject *args)
{
  unsigned short sts;
  GetLinkStatusResults gsr;
  PyObject * result;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }
  sbig_last_command = CC_GET_LINK_STATUS;
  sts = SBIGUnivDrvCommand(CC_GET_LINK_STATUS, NULL, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  result = Py_BuildValue("{s:i,s:i,s:i,s:i,s:i}", 
			 "established",      gsr.linkEstablished,
			 "lpt_base_address", gsr.baseAddress,
			 "type",             gsr.cameraType,
			 "total",            gsr.comTotal,
			 "failed",           gsr.comFailed);

  return result;
}

/* ----- set_driver_control ----------------------------------------------- *
 *
 * Modify the behavior of the driver by changing the setting 
 * of one of the driver control parameters 
 * [See DCP_* constants and section 3.5.12]
 * Be extra careful with that function !
 *
 */

static PyObject *
sbig_set_driver_control(PyObject *self, PyObject *args)
{
  unsigned short sts;
  SetDriverControlParams gsp;

  if (!PyArg_ParseTuple(args, "ii", 
			&gsp.controlParameter,
			&gsp.controlValue)) {
    return NULL;
  }
  sbig_last_command = CC_SET_DRIVER_CONTROL;
  sts = SBIGUnivDrvCommand(CC_SET_DRIVER_CONTROL, &gsp, NULL);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* ----- get_driver_control ----------------------------------------------- *
 *
 * Query setting of one of the driver control parameters
 *
 */

static PyObject *
sbig_get_driver_control(PyObject *self, PyObject *args)
{
  unsigned short sts;
  GetDriverControlParams gsp;
  GetDriverControlResults gsr;

  if (!PyArg_ParseTuple(args, "i", &gsp.controlParameter)) {
    return NULL;
  }
  sbig_last_command = CC_GET_DRIVER_CONTROL;
  sts = SBIGUnivDrvCommand(CC_GET_DRIVER_CONTROL, &gsp, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }
  return Py_BuildValue("i", gsr.controlValue);
}

/* ----- query_usb -------------------------------------------------------- *
 *
 * Look for USB cameras available on the USB bus.
 * This routine does not seem to work all the time. It seems to
 * only work once, and after no camera is found anymore. This
 * is not a big problem as it is still possible to use the other
 * routines, but it is just annoying.
 *
 */

static PyObject *
sbig_query_usb(PyObject *self, PyObject *args)
{
  unsigned short sts;
  QueryUSBResults gsr;
  const unsigned short device_address[4] = 
    { DEV_USB1, DEV_USB2, DEV_USB3, DEV_USB4 };
  PyObject * result;
  PyObject * item;
  int i;

  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  }

  sbig_last_command = CC_QUERY_USB;
  sts = SBIGUnivDrvCommand(CC_QUERY_USB, NULL, &gsr);
  if (sts != 0) {
    PyErr_SetString(sbig_error, sbig_error_string(sts));
    return NULL;
  }

  /* fprintf(stderr, "cameras found: %d\n", gsr.camerasFound); */

  if (gsr.camerasFound == 0) {
    return Py_BuildValue("()");
  }

  if (gsr.camerasFound > 4) {
    PyErr_SetString(sbig_error, 
		    "internal error: more than 4 USB devices found");
    return NULL;
  }    

  /* TODO: Check that this way to build the list does not imply 
     memory leaks... */

  result = Py_BuildValue("()");
  for(i=0 ; i<4 ; i++) {
    if (gsr.usbInfo[i].cameraFound == TRUE) {
      /* fprintf(stderr, "%p\n", result); */
      item = Py_BuildValue("({s:i,s:i,s:s,s:s})",
			   "device",  device_address[i],
			   "type",    gsr.usbInfo[i].cameraType,
			   "name",    gsr.usbInfo[i].name,
			   "serial",  gsr.usbInfo[i].serialNumber);
      result = PySequence_InPlaceConcat(result, item);

      Py_DECREF(item);  /* TODO: IS IT NEEDED ??? */
      /* fprintf(stderr, "%p\n", result); */
    }
  }

  return result;

}


/* ======================================================================== *
 * 
 * Declaration of all the defined methods for Python
 *
 */

static PyMethodDef sbig_methods[] = {
    /* -------------------------------------------------------------------- */
    {"open_driver",  sbig_open_driver, 
     METH_VARARGS, 
     "Open the driver."},
    {"close_driver",  sbig_close_driver, 
     METH_VARARGS, 
     "Close the driver."},
    {"open_device", (PyCFunction)sbig_open_device, 
     METH_VARARGS | METH_KEYWORDS, 
     "Load and initialize the low-level driver."},
    {"close_device",  sbig_close_device, 
     METH_VARARGS, 
     "Close the low-level driver."},
    {"get_driver_info", (PyCFunction)sbig_get_driver_info, 
     METH_VARARGS | METH_KEYWORDS, 
     "Return the version and capabilities of the driver."},
    {"get_driver_handle", sbig_get_driver_handle, 
     METH_VARARGS,
     "Return the handle of the current camera device (when using multiple cameras)."},
    {"set_driver_handle", (PyCFunction)sbig_set_driver_handle, 
     METH_VARARGS | METH_KEYWORDS, 
     "Set the current camera handle (when using multiple cameras)."},
    /* -------------------------------------------------------------------- */
    {"start_exposure", (PyCFunction)sbig_start_exposure, 
     METH_VARARGS | METH_KEYWORDS, 
     "initiate an exposure."},
    {"end_exposure", (PyCFunction)sbig_end_exposure, 
     METH_VARARGS | METH_KEYWORDS,
     "To use when the exposure is complete to prepare the CCD for readout, or to terminate the exposure prematurely."},
    {"start_readout", (PyCFunction)sbig_start_readout, 
     METH_VARARGS | METH_KEYWORDS,
     "Inform the driver about the area you intend to readout in subsequent readout_line calls."},
    {"readout_line", (PyCFunction)sbig_readout_line, 
     METH_VARARGS | METH_KEYWORDS,
     "Digitize some or all pixels in a row."},
    {"readout_lines", (PyCFunction)sbig_readout_lines, 
     METH_VARARGS | METH_KEYWORDS,
     "Digitize some or all pixels in several rows at once."},
    {"dump_lines", (PyCFunction)sbig_dump_lines, 
     METH_VARARGS | METH_KEYWORDS,
     "Discard entire rows (for partial readout)."},
    {"end_readout", (PyCFunction)sbig_end_readout, 
     METH_VARARGS | METH_KEYWORDS,
     "Prepare the CCD for idle state after readout."},
    /* -------------------------------------------------------------------- */
    {"query_temperature_status", sbig_query_temperature_status,
     METH_VARARGS,
     "Return the temperature regulation parameters."},
    {"set_temperature_regulation", 
     (PyCFunction)sbig_set_temperature_regulation,
     METH_VARARGS | METH_KEYWORDS,
     "Enable or disable the CCD temperature regulation."},
    /* -------------------------------------------------------------------- */
    {"establish_link", sbig_establish_link,
     METH_VARARGS,
     "Establish the link with the CCD camera."},
    {"get_ccd_info", (PyCFunction)sbig_get_ccd_info,
     METH_VARARGS | METH_KEYWORDS,
     "Return camera informations and capabilities"},
    {"query_command_status", (PyCFunction)sbig_query_command_status,
     METH_VARARGS | METH_KEYWORDS,
     "Status of the last command (See STATUS_* constants)."},
    {"miscellaneous_control", (PyCFunction)sbig_miscellaneous_control,
     METH_VARARGS | METH_KEYWORDS,
     "Control the Fan, LED and shutter"},
    {"get_link_status", sbig_get_link_status,
     METH_VARARGS,
     "Returns the status of the link established with the camera."},
    {"read_offset", (PyCFunction)sbig_read_offset,
     METH_VARARGS | METH_KEYWORDS,
     "Returns the CCD offset."},
    {"set_driver_control", sbig_set_driver_control,
     METH_VARARGS,
     "Modify the behavior of the driver by changing the setting of one of the driver control parameters (See DCP_* constants)."},
    {"get_driver_control", sbig_get_driver_control,
     METH_VARARGS,
     "Query the setting of one of the driver control parameters."},
    {"query_usb", sbig_query_usb,
     METH_VARARGS,
     "Query the USB bus and detect up to 4 USB cameras."},
    /* -------------------------------------------------------------------- */
    {NULL, NULL, 0, NULL}        /* Sentinel */
    /* -------------------------------------------------------------------- */
};

/* ======================================================================== *
 * 
 * Module initialization function
 *
 */

PyMODINIT_FUNC
initsbig(void)
{
  PyObject *_module;
  _module = Py_InitModule("sbig", sbig_methods);

  /* ----- Numarray initialization ---------------------------------------- */

  import_libnumarray();

  /* ----- Defining a specific Exception ---------------------------------- */

  sbig_error = PyErr_NewException("sbig.SBIGDriverError", NULL, NULL);
  Py_INCREF(sbig_error);
  PyModule_AddObject(_module, "SBIGDriverError", sbig_error);

  /* ----- Defining some constants in Python namespace -------------------- */

  /* ----- __version__ constant ------------------ */

  PyModule_AddObject(_module, "__version__", Py_BuildValue("s", VERSION));

  /* ----- Devices available --------------------- */

  PyModule_AddObject(_module, "DEV_NONE", PyInt_FromLong(DEV_NONE));
  PyModule_AddObject(_module, "DEV_LPT1", PyInt_FromLong(DEV_LPT1));
  PyModule_AddObject(_module, "DEV_LPT2", PyInt_FromLong(DEV_LPT2));
  PyModule_AddObject(_module, "DEV_LPT3", PyInt_FromLong(DEV_LPT3));
  PyModule_AddObject(_module, "DEV_USB",  PyInt_FromLong(DEV_USB));
  PyModule_AddObject(_module, "DEV_ETH",  PyInt_FromLong(DEV_ETH));
  PyModule_AddObject(_module, "DEV_USB1", PyInt_FromLong(DEV_USB1));
  PyModule_AddObject(_module, "DEV_USB2", PyInt_FromLong(DEV_USB2));
  PyModule_AddObject(_module, "DEV_USB3", PyInt_FromLong(DEV_USB3));
  PyModule_AddObject(_module, "DEV_USB4", PyInt_FromLong(DEV_USB4));

  /* ----- CCD used ------------------------------ */

  PyModule_AddObject(_module, "CCD_IMAGING", 
		     PyInt_FromLong(CCD_IMAGING));
  PyModule_AddObject(_module, "CCD_TRACKING", 
		     PyInt_FromLong(CCD_TRACKING));
  PyModule_AddObject(_module, "CCD_EXT_TRACKING", 
		     PyInt_FromLong(CCD_EXT_TRACKING));

  /* ----- Camera types -------------------------- */

  PyModule_AddObject(_module, "ST7_CAMERA",   PyInt_FromLong(ST7_CAMERA));
  PyModule_AddObject(_module, "ST8_CAMERA",   PyInt_FromLong(ST8_CAMERA));
  PyModule_AddObject(_module, "ST5C_CAMERA",  PyInt_FromLong(ST5C_CAMERA));
  PyModule_AddObject(_module, "TCE_CONTROLLER",
		     PyInt_FromLong(TCE_CONTROLLER));
  PyModule_AddObject(_module, "ST237_CAMERA", PyInt_FromLong(ST237_CAMERA));
  PyModule_AddObject(_module, "STK_CAMERA",   PyInt_FromLong(STK_CAMERA));
  PyModule_AddObject(_module, "ST9_CAMERA",   PyInt_FromLong(ST9_CAMERA));
  PyModule_AddObject(_module, "STV_CAMERA",   PyInt_FromLong(STV_CAMERA));
  PyModule_AddObject(_module, "ST10_CAMERA",  PyInt_FromLong(ST10_CAMERA));
  PyModule_AddObject(_module, "ST1K_CAMERA",  PyInt_FromLong(ST1K_CAMERA));
  PyModule_AddObject(_module, "ST2K_CAMERA",  PyInt_FromLong(ST2K_CAMERA));
  PyModule_AddObject(_module, "STL_CAMERA",   PyInt_FromLong(STL_CAMERA));
  PyModule_AddObject(_module, "ST402_CAMERA", PyInt_FromLong(ST402_CAMERA));
  PyModule_AddObject(_module, "NEXT_CAMERA",  PyInt_FromLong(NEXT_CAMERA));
  PyModule_AddObject(_module, "NO_CAMERA",    PyInt_FromLong(NO_CAMERA));

  /* ----- query_command_status status codes ----- */

  PyModule_AddObject(_module, "STATUS_IDLE",  
		     PyInt_FromLong(CS_IDLE));
  PyModule_AddObject(_module, "STATUS_IN_PROGRESS",  
		     PyInt_FromLong(CS_IN_PROGRESS));
  PyModule_AddObject(_module, "STATUS_INTEGRATING",  
		     PyInt_FromLong(CS_INTEGRATING));
  PyModule_AddObject(_module, "STATUS_COMPLETE",  
		     PyInt_FromLong(CS_INTEGRATION_COMPLETE));

  /* ----- driver control parameters ------------- */

  PyModule_AddObject(_module, "DCP_USB_FIFO_ENABLE",  
		     PyInt_FromLong(DCP_USB_FIFO_ENABLE));
  PyModule_AddObject(_module, "DCP_CALL_JOURNAL_ENABLE",  
		     PyInt_FromLong(DCP_CALL_JOURNAL_ENABLE));
  PyModule_AddObject(_module, "DCP_IVTOH_RATIO",  
		     PyInt_FromLong(DCP_IVTOH_RATIO));
  PyModule_AddObject(_module, "DCP_USB_FIFO_SIZE",  
		     PyInt_FromLong(DCP_USB_FIFO_SIZE));
  PyModule_AddObject(_module, "DCP_USB_DRIVER",  
		     PyInt_FromLong(DCP_USB_DRIVER));
  PyModule_AddObject(_module, "DCP_KAI_RELGAIN",  
		     PyInt_FromLong(DCP_KAI_RELGAIN));
  PyModule_AddObject(_module, "DCP_USB_PIXEL_DL_ENABLE",  
		     PyInt_FromLong(DCP_USB_PIXEL_DL_ENABLE));
  PyModule_AddObject(_module, "DCP_HIGH_THROUGHPUT",  
		     PyInt_FromLong(DCP_HIGH_THROUGHPUT));
  PyModule_AddObject(_module, "DCP_VDD_OPTIMIZED",  
		     PyInt_FromLong(DCP_VDD_OPTIMIZED));
  PyModule_AddObject(_module, "DCP_AUTO_AD_GAIN",  
		     PyInt_FromLong(DCP_AUTO_AD_GAIN));
  PyModule_AddObject(_module, "DCP_LAST",  
		     PyInt_FromLong(DCP_LAST));

  /* ----- temperature regulation modes ---------- */

  PyModule_AddObject(_module, "REGULATION_OFF",  
		     PyInt_FromLong(REGULATION_OFF));
  PyModule_AddObject(_module, "REGULATION_ON",  
		     PyInt_FromLong(REGULATION_ON));
  PyModule_AddObject(_module, "REGULATION_OVERRIDE",  
		     PyInt_FromLong(REGULATION_OVERRIDE));
  PyModule_AddObject(_module, "REGULATION_FREEZE",  
		     PyInt_FromLong(REGULATION_FREEZE));
  PyModule_AddObject(_module, "REGULATION_UNFREEZE",  
		     PyInt_FromLong(REGULATION_UNFREEZE));
  PyModule_AddObject(_module, "REGULATION_ENABLE_AUTOFREEZE",  
		     PyInt_FromLong(REGULATION_ENABLE_AUTOFREEZE));
  PyModule_AddObject(_module, "REGULATION_DISABLE_AUTOFREEZE",  
		     PyInt_FromLong(REGULATION_DISABLE_AUTOFREEZE));

  /* ----- CCD info requests --------------------- */

  PyModule_AddObject(_module, "CCD_INFO_IMAGING",  
		     PyInt_FromLong(CCD_INFO_IMAGING));
  PyModule_AddObject(_module, "CCD_INFO_TRACKING",  
		     PyInt_FromLong(CCD_INFO_TRACKING));
  PyModule_AddObject(_module, "CCD_INFO_EXTENDED",  
		     PyInt_FromLong(CCD_INFO_EXTENDED));
  PyModule_AddObject(_module, "CCD_INFO_EXTENDED_5C",  
		     PyInt_FromLong(CCD_INFO_EXTENDED_5C));
  PyModule_AddObject(_module, "CCD_INFO_EXTENDED2_IMAGING",  
		     PyInt_FromLong(CCD_INFO_EXTENDED2_IMAGING));
  PyModule_AddObject(_module, "CCD_INFO_EXTENDED2_TRACKING",  
		     PyInt_FromLong(CCD_INFO_EXTENDED2_TRACKING));

  /* ----- CCD shutter status -------------------- */

  PyModule_AddObject(_module, "SC_LEAVE_SHUTTER",  
		     PyInt_FromLong(SC_LEAVE_SHUTTER));
  PyModule_AddObject(_module, "SC_OPEN_SHUTTER",  
		     PyInt_FromLong(SC_OPEN_SHUTTER));
  PyModule_AddObject(_module, "SC_CLOSE_SHUTTER",  
		     PyInt_FromLong(SC_CLOSE_SHUTTER));
  PyModule_AddObject(_module, "SC_INITIALIZE_SHUTTER",  
		     PyInt_FromLong(SC_INITIALIZE_SHUTTER));
  PyModule_AddObject(_module, "SC_OPEN_EXT_SHUTTER",  
		     PyInt_FromLong(SC_OPEN_EXT_SHUTTER));
  PyModule_AddObject(_module, "SC_CLOSE_EXT_SHUTTER",  
		     PyInt_FromLong(SC_CLOSE_EXT_SHUTTER));

  /* ----- LED behaviour ------------------------- */

  PyModule_AddObject(_module,  "LED_OFF", 
		     PyInt_FromLong(LED_OFF));
  PyModule_AddObject(_module,  "LED_ON", 
		     PyInt_FromLong(LED_ON));
  PyModule_AddObject(_module,  "LED_BLINK_LOW", 
		     PyInt_FromLong(LED_BLINK_LOW));
  PyModule_AddObject(_module,  "LED_BLINK_HIGH", 
		     PyInt_FromLong(LED_BLINK_HIGH));

}

/* ======================================================================== */
